# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.scan_paths"""

import gc
import json
import logging
import os
import threading
import time
import uuid

from datetime import datetime
from multiprocessing import cpu_count

import six

from boto3.s3 import transfer as S3Transfer

from sonicprobe.libs import workerpool

from .common import (DATE_FORMAT, STATUS_PROCESSING, STATUS_STOPPED)
from .s3 import (XtrargonTxProgress, XTRARGON_FILE)


LOG                         = logging.getLogger('xtrargon_ng.classes.scan_paths')

DEFAULT_MAX_WORKERS         = 'auto'
DEFAULT_MAX_LIFETIME        = 300
DEFAULT_MAX_TASKS           = 100
DEFAULT_MAX_UPLOAD_ATTEMPTS = 10
DEFAULT_OBJECT_TTL          = 604800
DEFAULT_STATE_FILE_FORMAT   = "/{server_id}/{backup_type}/{time:%Y-%m-%dT%H:%M:%S}/xtrargon.state"


class BackupsScanPaths(threading.Thread):
    def __init__(self, module):
        threading.Thread.__init__(self)
        self.name            = 'scan'
        self._module         = module
        self._config         = module.config
        self._s3helper       = module.s3helper

        self._step           = 1
        self._max_attempts   = 1
        self._killed         = False

        self._bucket         = self._s3helper.get_bucket()
        self._storage_class  = self._s3helper.get_storage_class()
        self._client         = self._s3helper.boto3_client()
        self._object_ttl     = DEFAULT_OBJECT_TTL
        self._file_sync      = module.file_sync

        self._backup_type    = 'full'

        self._uploaders      = None
        self._upload_files   = []
        self._fcache         = {}

        if 'object_ttl' in self._config['backups']:
            self._object_ttl = int(self._config['backups']['object_ttl'])

    def _load_uploaders(self):
        max_workers     = self._config['backups'].get('max_workers', DEFAULT_MAX_WORKERS) or 1
        if max_workers == 'auto':
            max_workers = cpu_count()
        elif max_workers < 1:
            max_workers = 1

        worker_lifetime= int(self._config['backups'].get('worker_lifetime', DEFAULT_MAX_LIFETIME))
        max_tasks       = int(self._config['backups'].get('max_tasks', DEFAULT_MAX_TASKS))
        self._uploaders = workerpool.WorkerPool(max_workers = int(max_workers),
                                                life_time   = worker_lifetime,
                                                max_tasks   = max_tasks,
                                                name        = 'uploaders',
                                                auto_gc     = False)

    @staticmethod
    def _mk_sum(filepath):
        if not os.path.exists(filepath):
            return ''

        fstat = os.stat(filepath)

        return '|'.join((str(fstat.st_size),
                         str(fstat.st_mtime)))

    def _mk_filekey(self, filepath, xid = None):
        if xid:
            return "%s:%s:file:%s" % (self._module.server_id, xid, filepath)

        return "%s:%s:file:%s" % (self._module.server_id, xid, filepath)

    def _chk_fcache(self, toppath, srcpath, fkey, nsum):
        if toppath not in self._fcache:
            self._fcache[toppath] = {}

        if srcpath in self._fcache[toppath]:
            osum = self._fcache[toppath][srcpath]
            if osum and osum == nsum:
                return True

        osum = self._file_sync.get_key(fkey)
        if osum and osum == nsum:
            return True

        return False

    def _s3up(self, src, dst):
        if not dst:
            dst = src

        if not os.path.exists(src):
            return (None, dst)

        last_exception = None

        max_attempts = 1
        if self._step > 1:
            max_attempts = self._max_attempts

        for i in range(1, max_attempts + 1):
            try:
                with S3Transfer.S3Transfer(self._s3helper.boto3_client(),
                                           self._s3helper.transfer_config()) as transfer:
                    transfer.upload_file(
                        filename   = src,
                        bucket     = self._bucket,
                        key        = dst.lstrip('/'),
                        callback   = XtrargonTxProgress(src, False, self._killed),
                        extra_args = {'StorageClass': self._storage_class})

                LOG.info("s3 upload (src: %r, dst: %r, bucket: %r, attempt: %s/%s)",
                         src,
                         dst,
                         self._bucket,
                         i,
                         max_attempts)
                return (True, dst)
            except GeneratorExit as e:
                return (GeneratorExit(), dst)
            except Exception as e:
                LOG.debug("Retrying exception caught (%s), "
                          "retrying request, (attempt %s/%s)", e, i,
                          max_attempts, exc_info=True)
                last_exception = e
                continue

        if self._step > 1:
            return (Exception(last_exception), dst)

        return (None, dst)

    def _s3up_complete(self, ret):
        result, destpath = ret

        if destpath and destpath in self._upload_files:
            self._upload_files.remove(destpath)

        if isinstance(result, GeneratorExit):
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        elif isinstance(result, Exception):
            self._module.add_errors(repr(result))
            LOG.exception(result)

    def _s3api(self, cmd, key = None, **kwargs):
        kwargs['Bucket'] = self._bucket
        if key:
            kwargs['Key'] = key.lstrip('/')
        kwargs['StorageClass'] = self._storage_class

        LOG.debug("s3 api: %r: %r", cmd, kwargs)
        getattr(self._client, cmd)(**kwargs)

    def _put_file(self, xpath, body = ''):
        self._s3api('put_object', xpath, **{'Body': body})

    def _pre_upload(self, toppath, srcpath, force = False):
        xid  = self._module.get_state_value('started_at')
        fkey = self._mk_filekey(srcpath, xid)
        nsum = self._mk_sum(srcpath)

        if not force and self._chk_fcache(toppath, srcpath, fkey, nsum):
            return False

        return (fkey, nsum)

    def _post_upload(self, toppath, srcpath, fkey, nsum):
        if toppath not in self._fcache:
            self._fcache[toppath] = {}
        self._fcache[toppath][srcpath] = nsum

        self._file_sync.set_key(fkey, nsum, self._object_ttl)

    @staticmethod
    def _destpath(toppath, params, srcpath, xvars):
        destpath = srcpath
        fkwargs  = xvars.copy()
        fkwargs['filepath'] = srcpath
        fkwargs['dirname']  = os.path.dirname(srcpath)
        fkwargs['filename'] = os.path.basename(srcpath)

        if not params or not isinstance(params, dict):
            return destpath

        if params.get('format'):
            destpath = params['format'].format(**fkwargs)

        if params.get('subst'):
            if isinstance(params['subst'], six.string_types):
                xsubst = params['subst']
            else:
                xsubst = toppath

            if srcpath.startswith(xsubst):
                substpath = srcpath[len(xsubst):].strip(os.path.sep)
                if params.get('format'):
                    destpath = os.path.join(os.path.sep, destpath.strip(os.path.sep), substpath)
                else:
                    destpath = os.path.join(os.path.sep, substpath)

        if params.get('prefix'):
            destpath = os.path.join(params['prefix'].strip(os.path.sep),
                                    destpath.strip(os.path.sep))

        return os.path.join(os.path.sep, destpath.strip(os.path.sep))

    def _set_max_attempts(self, params):
        self._max_attempts = int(params.get('max_upload_attempts', DEFAULT_MAX_UPLOAD_ATTEMPTS) or 1)
        if self._max_attempts < 1:
            self._max_attempts = 1

    def _walk_from_cache(self, toppath, params, xvars):
        LOG.info("walk from cache")

        fcache = self._fcache.copy()

        for srcpath in six.iterkeys(fcache[toppath]):
            if self._killed:
                raise GeneratorExit()

            self._set_max_attempts(params)

            destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

            res = self._pre_upload(toppath, srcpath)
            if not res:
                continue

            self._upload_files.append(destpath)
            self._uploaders.run(self._s3up,
                                complete = self._s3up_complete,
                                src      = srcpath,
                                dst      = destpath)

            self._post_upload(toppath, srcpath, res[0], res[1])

    def _walk(self, toppath, params, xvars):
        LOG.info("walk path: %r", toppath)

        for root, dirs, files in os.walk(toppath, topdown = True, followlinks = params['follow']):
            if self._killed:
                raise GeneratorExit()

            if params['exclude_patterns'] and params['exclude_patterns'](root):
                dirs[:] = []
                LOG.debug("exclude path from scan. (path: %r)", root)
                continue

            self._set_max_attempts(params)

            if not dirs and not files:
                srcpath  = os.path.join(root, XTRARGON_FILE)
                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)
                self._put_file(destpath)
                continue

            for filename in files:
                if self._killed:
                    raise GeneratorExit()

                srcpath  = os.path.join(root, filename)

                if params['exclude_patterns'] and params['exclude_patterns'](srcpath):
                    LOG.debug("exclude file from scan. (filepath: %r)", srcpath)
                    continue

                destpath = self._destpath(toppath, params.get('dest_params'), srcpath, xvars)

                res = self._pre_upload(toppath, srcpath)
                if not res:
                    continue

                self._upload_files.append(destpath)
                self._uploaders.run(self._s3up,
                                    complete = self._s3up_complete,
                                    src      = srcpath,
                                    dst      = destpath)

                self._post_upload(toppath, srcpath, res[0], res[1])

    def _run(self, fkwargs):
        for toppath, params in six.iteritems(self._config['backups']['paths']):
            if self._killed:
                raise GeneratorExit()

            if self._step == 1 \
               or not bool(params.get('second_step_from_cache', True)) \
               or toppath not in self._fcache:
                self._walk(toppath, params, fkwargs)
            else:
                self._walk_from_cache(toppath, params, fkwargs)

        while self._upload_files:
            LOG.info("!!!! TODO add a timeout !!!!")
            LOG.info("upload files missing: %r", self._upload_files)
            if self._killed:
                raise GeneratorExit()

            time.sleep(0.5)

    def _post_run(self, fkwargs):
        LOG.info("finalizing backup")

        self._fcache = {}

        try:
            state = self._module.get_state(end = True)
            state_file_format = self._config['backups'].get('state_file_format') or DEFAULT_STATE_FILE_FORMAT
            self._put_file(state_file_format.format(**fkwargs),
                           body = json.dumps(state))

            if self._config['backups'].get('notifications'):
                self._module.notifiers(state)

            if self._uploaders:
                self._uploaders.killall(1)
                self._uploaders = None

            gc.collect()
        except Exception as e:
            LOG.exception(e)

    def run(self):
        self._module.is_running.wait()
        LOG.info("starting backup")
        self._module.set_state_value('started_at', time.strftime(DATE_FORMAT))
        self._module.set_state_value('uuid', "%s" % uuid.uuid4())
        self._module.set_state_value('status', STATUS_PROCESSING)
        self._module.is_scanning.set()
        self._fcache = {}

        fkwargs = {'env': os.environ,
                   'backup_type': self._backup_type,
                   'gmtime': datetime.utcnow(),
                   'time': datetime.now(),
                   'server_id': self._module.server_id}

        try:
            self._load_uploaders()

            if self._killed:
                raise GeneratorExit()

            LOG.info("first step of backup")
            self._step = 1
            self._run(fkwargs)
            self._module.lock_db.set()
            self._module.db_locked.wait()

            if self._killed:
                raise GeneratorExit()

            LOG.info("second step of backup")
            self._step = 2
            self._run(fkwargs)
        except GeneratorExit:
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        except Exception as e:
            self._module.add_errors(repr(e))
            LOG.exception(e)
        finally:
            self._module.lock_db.clear()
            self._module.is_scanning.clear()
            self._module.is_running.clear()

        self._post_run(fkwargs)

    def terminate(self):
        LOG.info("backup terminated")
        self._killed = True
        self._module.is_running.set()
        self._module.is_scanning.set()

        if self._uploaders:
            self._uploaders.killall(1)
